@@ -49,7 +49,8 @@ static void acm_set_server_port(void)
 	FILE *f;
 
 	if ((f = fopen("/var/run/ibacm.port", "r"))) {
-		fscanf(f, "%hu", (unsigned short *) &server_port);
+		if (fscanf(f, "%hu", (unsigned short *) &server_port) != 1) 
+			printf("Failed to read server port\n"); 
 		fclose(f);
 	}
 }
@@ -304,7 +305,7 @@ out:
 	return ret;
 }
 
-int ib_acm_query_perf(uint64_t **counters, int *count)
+int ib_acm_query_perf(int index, uint64_t **counters, int *count)
 {
 	struct acm_msg msg;
 	int ret, i;
@@ -313,6 +314,7 @@ int ib_acm_query_perf(uint64_t **counters, int *count)
 	memset(&msg, 0, sizeof msg);
 	msg.hdr.version = ACM_VERSION;
 	msg.hdr.opcode = ACM_OP_PERF_QUERY;
+	msg.hdr.data[1] = index;
 	msg.hdr.length = htons(ACM_MSG_HDR_LENGTH);
 
 	ret = send(sock, (char *) &msg, ACM_MSG_HDR_LENGTH, 0);
@@ -345,6 +347,110 @@ out:
 	return ret;
 }
 
+int ib_acm_enum_ep(int index, struct acm_ep_config_data **data)
+{
+	struct acm_msg msg;
+	int ret;
+	int len;
+	int cnt;
+	struct acm_ep_config_data *edata;
+
+	lock_acquire(&lock);
+	memset(&msg, 0, sizeof msg);
+	msg.hdr.version = ACM_VERSION;
+	msg.hdr.opcode = ACM_OP_EP_QUERY;
+	msg.hdr.data[0] = index;
+	msg.hdr.length = htons(ACM_MSG_HDR_LENGTH);
+
+	ret = send(sock, (char *) &msg, ACM_MSG_HDR_LENGTH, 0);
+	if (ret != ACM_MSG_HDR_LENGTH)
+		goto out;
+
+	ret = recv(sock, (char *) &msg, sizeof msg, 0);
+	if (ret < ACM_MSG_HDR_LENGTH || ret != ntohs(msg.hdr.length)) {
+		ret = ACM_STATUS_EINVAL;
+		goto out;
+	}
+
+	if (msg.hdr.status) {
+		ret = acm_error(msg.hdr.status);
+		goto out;
+	}
+
+	cnt = ntohs(msg.ep_data[0].addr_cnt);
+	len = sizeof(struct acm_ep_config_data) + 
+		ACM_MAX_ADDRESS * cnt;
+	edata = malloc(len);
+	if (!edata) {
+		ret = ACM_STATUS_ENOMEM;
+		goto out;
+	}
+
+	memcpy(edata, &msg.ep_data[0], len);
+	edata->dev_guid = ntohll(msg.ep_data[0].dev_guid);
+	edata->pkey = ntohs(msg.ep_data[0].pkey);
+	edata->addr_cnt = cnt;
+	*data = edata;
+	ret = 0;
+out:
+	lock_release(&lock);
+	return ret;
+}
+
+int ib_acm_query_perf_ep_addr(uint8_t *src, uint8_t type, 
+			     uint64_t **counters, int *count)
+{
+	struct acm_msg msg;
+	int ret, i, len;
+
+	if (!src) 
+		return -1;
+
+	lock_acquire(&lock);
+	memset(&msg, 0, sizeof msg);
+	msg.hdr.version = ACM_VERSION;
+	msg.hdr.opcode = ACM_OP_PERF_QUERY;
+
+	ret = acm_format_ep_addr(&msg.resolve_data[0], src, type,
+		ACM_EP_FLAG_SOURCE);
+	if (ret)
+		goto out;
+
+	len = ACM_MSG_HDR_LENGTH + ACM_MSG_EP_LENGTH;
+	msg.hdr.length = htons(len);
+
+	ret = send(sock, (char *) &msg, len, 0);
+	if (ret != len)
+		goto out;
+
+	ret = recv(sock, (char *) &msg, sizeof msg, 0);
+	if (ret < ACM_MSG_HDR_LENGTH || ret != ntohs(msg.hdr.length)) {
+		ret = ACM_STATUS_EINVAL;
+		goto out;
+	}
+
+	if (msg.hdr.status) {
+		ret = acm_error(msg.hdr.status);
+		goto out;
+	}
+
+	*counters = malloc(sizeof(uint64_t) * msg.hdr.data[0]);
+	if (!*counters) {
+		ret = ACM_STATUS_ENOMEM;
+		goto out;
+	}
+
+	*count = msg.hdr.data[0];
+	for (i = 0; i < *count; i++)
+		(*counters)[i] = ntohll(msg.perf_data[i]);
+
+	ret = 0;
+out:
+	lock_release(&lock);
+	return ret;
+}
+
+
 const char *ib_acm_cntr_name(int index)
 {
 	static const char *const cntr_name[] = {
